import os
import yaml
import shutil
import functools
import numpy as np

from . import core
from . import utils
from . import io_utils
from . import wfunc_utils

from .Catalog import Catalog
from .Cosmology import Cosmology
from .SurrogateFactory import SurrogateFactory


class KszPipe:
    def __init__(self, input_dir, output_dir):
        r"""This is the main kSZ analysis pipeline, which computes data/surrogate power spectra from catalogs.

        There are two ways to run a KszPipe pipeline. The first way is to create a KszPipe instance and
        call the ``run()`` method. This can be done either in a script or a jupyter notebook (if you're
        using jupyter, you should keep in mind that the KszPipe may take hours to run, and you'll need to
        babysit the connection to the jupyterhib). The second way is to run from the command line with::
        
           python -m kszx kszpipe_run [-p NUM_PROCESSES] <input_dir> <output_dir>
        
        The ``input_dir`` contains a parameter file ``params.yml`` and galaxy/random catalogs.
        The ``output_dir`` will be populated with power spectra.
        For details of what KszPipe computes, and documentation of file formats, see the sphinx docs:

          https://kszx.readthedocs.io/en/latest/kszpipe.html#kszpipe-details

        After running the pipeline, you may want to load pipeline outputs using the helper class
        :class:`~kszx.KszPipeOutdir`, or do parameter estimation using :class:`~kszx.PgvLikelihood`.
        
        High-level features:
        
          - Runs "surrogate" sims (see overleaf) to characterize the survey window function,
            determine dependence of power spectra on $(f_{NL}, b_v)$, and assign error bars
            to power spectra.
 
          - Velocity reconstruction noise is included in surrogate sims via a bootstrap procedure,
            using the observed CMB realization. This automatically incorporates noise inhomogeneity
            and "striping", and captures correlations e.g. between 90 and 150 GHz.

          - The galaxy catalog can be spectroscopic or photometric (via the ``ztrue_col`` and 
            ``zobs_col`` constructor args). Surrogate sims will capture the effect of photo-z errors.

          - The windowed power spectra $P_{gg}$, $P_{gv}$, $P_{vv}$ use a normalization which
            should be approximately correct. The normalization is an ansatz which is imperfect,
            especially on large scales, so surrogate sims should still be used to compare power
            spetra to models. Eventually, we'll implement a precise calculation of the window
            function.

          - Currently assumes one galaxy field, and two velocity reconstructions labelled
            "90" and "150" (with ACT in mind).

          - Currently, there is not much implemented for CMB foregrounds. Later, I'd like
            to include foreground clustering terms in the surrogate model (i.e. terms of the
            form $b_\delta \delta(x)$, in addition to the kSZ term $b_v v_r(x)$), and estimate
            the $b_\delta$ biases by estimating the spin-zero $P_{gv}$ power spectrum.
        """
        
        self.input_dir = input_dir
        self.output_dir = output_dir
        
        # Create output_dir and 'tmp' subdir
        os.makedirs(f'{output_dir}/tmp', exist_ok=True)

        # Read files from pipeline input_dir.
        # Reference: https://kszx.readthedocs.io/en/latest/kszpipe.html#kszpipe-details
        
        with open(f'{input_dir}/params.yml', 'r') as f:
            params = yaml.safe_load(f)

            self.version = params['version']
            assert self.version == 1   # for now

            self.nsurr = params['nsurr']
            self.surr_bg = params['surr_bg']
            self.nzbins_gal = params['nzbins_gal']
            self.nzbins_vr = params['nzbins_vr']

            self.kmax = params['kmax']
            self.nkbins = params['nkbins']            
            self.kbin_edges = np.linspace(0, self.kmax, self.nkbins+1)
            self.kbin_centers = (self.kbin_edges[1:] + self.kbin_edges[:-1]) / 2.

        self.box = io_utils.read_pickle(f'{input_dir}/bounding_box.pkl')
        self.kernel = 'cubic'   # hardcoded for now
        self.deltac = 1.68      # hardcoded for now

        self.pk_data_filename = f'{output_dir}/pk_data.npy'
        self.pk_surr_filename = f'{output_dir}/pk_surrogates.npy'
        self.pk_single_surr_filenames = [ f'{output_dir}/tmp/pk_surr_{i}.npy' for i in range(self.nsurr) ]


    @functools.cached_property
    def cosmo(self):
        return Cosmology('planck18+bao')

    @functools.cached_property
    def gcat(self):
        return Catalog.from_h5(f'{self.input_dir}/galaxies.h5')

    @functools.cached_property
    def rcat(self):
        return Catalog.from_h5(f'{self.input_dir}/randoms.h5')

    @functools.cached_property
    def rcat_xyz_obs(self):
        return self.rcat.get_xyz(self.cosmo, 'zobs')

    @functools.cached_property
    def sum_rcat_gweights(self):
        return np.sum(self.rcat.weight_gal) if hasattr(self.rcat, 'weight_gal') else self.rcat.size

    @functools.cached_property
    def sum_rcat_vr_weights(self):
        return np.sum(self.rcat.weight_vr) if hasattr(self.rcat, 'weight_vr') else self.rcat.size

    
    @functools.cached_property
    def window_function(self):
        r"""
        3-by-3 matrix $W_{ij}$ containing the window function for power spectra on three spatial footprints:
        
          - footprint 0: random catalog weighted by ``weight_gal`` column
          - footprint 1: random catalog weighted by product of columns ``weight_vr * bv_90``
          - footprint 2: random catalog weighted by product of columns ``weight_vr * bv_150``

        These spatial weightings are appropriate for the $\delta_g$, $v_r^{90}$, and $v_r^{150}$ fields.

        Window functions are computed with ``wfunc_utils.compute_wcrude()`` and are crude approximations
        (for more info see :func:`~kszx.wfunc_utils.compute_wcrude()` docstring), but this is okay since
        surrogate fields are treated consistently.
        """
        
        print('Initializing KszPipe.window_function')
        
        nrand = self.rcat.size
        rweights = getattr(self.rcat, 'weight_gal', np.ones(nrand))
        vweights = getattr(self.rcat, 'weight_vr', np.ones(nrand))

        # Cache these properties for later use (not logically necessary, but may help later pipeline stages run faster)
        self.sum_rcat_gweights
        self.sum_rcat_vr_weights
        
        # Fourier-space maps representing footprints.
        footprints = [
            core.grid_points(self.box, self.rcat_xyz_obs, rweights, kernel=self.kernel, fft=True, compensate=True),
            core.grid_points(self.box, self.rcat_xyz_obs, vweights * self.rcat.bv_90, kernel=self.kernel, fft=True, compensate=True),
            core.grid_points(self.box, self.rcat_xyz_obs, vweights * self.rcat.bv_150, kernel=self.kernel, fft=True, compensate=True)
        ]

        # Compute window function using wfunc_utils.compute_wcrude().
        wf = wfunc_utils.compute_wcrude(self.box, footprints)
        
        print('KszPipe.window_function initialized')
        return wf

    
    @functools.cached_property
    def surrogate_factory(self):
        """
        Returns an instance of class SurrogateFactory, a helper class for simulating the
        density and radial velocity fields at locations of randoms.
        """
        
        print('Initializing KszPipe.surrogate_factory')

        surr_ngal_mean = self.gcat.size
        surr_ngal_rms = 4 * np.sqrt(self.gcat.size)  # 4x Poisson        
        sf = SurrogateFactory(self.box, self.cosmo, self.rcat, surr_ngal_mean, surr_ngal_rms, 'ztrue')

        print('KszPipe.surrogate_factory initialized')
        return sf

    
    def get_pk_data(self, run=False, force=False):
        r"""Returns a shape (3,3,nkbins) array, and saves it in ``pipeline_outdir/pk_data.npy``.

        The returned array contains auto and cross power spectra of the following fields:

          - 0: galaxy overdensity 
          - 1: kSZ velocity reconstruction $v_r^{90}$
          - 2: kSZ velocity reconstruction $v_r^{150}$

        Flags:
        
          - If ``run=False``, then this function expects the $P(k)$ file to be on disk from a previous pipeline run.
          - If ``run=True``, then the $P(k)$ file will be computed if it is not on disk.
          - If ``force=True``, then this function recomputes $P(k)$, even if it is on disk from a previous pipeline run.
        """
        
        if (not force) and os.path.exists(self.pk_data_filename):
            return io_utils.read_npy(self.pk_data_filename)

        if not (run or force):
            raise RuntimeError(f'KszPipe.get_pk_data2(): run=force=False was specified, and file {self.pk_data_filename} not found')
        
        print('get_pk_data2(): running\n', end='')
        
        gweights = getattr(self.gcat, 'weight_gal', np.ones(self.gcat.size))
        rweights = getattr(self.rcat, 'weight_gal', np.ones(self.rcat.size))
        vweights = getattr(self.gcat, 'weight_vr', np.ones(self.gcat.size))
        gcat_xyz = self.gcat.get_xyz(self.cosmo)

        # To mitigate CMB foregrounds, we apply mean-subtraction to the vr fields.
        # (Note that we perform the same mean-subtraction to surrogate fields, in get_pk_surrogate().)
        
        coeffs_v90 = utils.subtract_binned_means(vweights * self.gcat.tcmb_90, self.gcat.z, self.nzbins_vr)
        coeffs_v150 = utils.subtract_binned_means(vweights * self.gcat.tcmb_150, self.gcat.z, self.nzbins_vr)

        # Note spin=1 on the vr FFTs.
        fourier_space_maps = [
            core.grid_points(self.box, gcat_xyz, gweights, self.rcat_xyz_obs, rweights, kernel=self.kernel, fft=True, compensate=True),
            core.grid_points(self.box, gcat_xyz, coeffs_v90, kernel=self.kernel, fft=True, spin=1, compensate=True),
            core.grid_points(self.box, gcat_xyz, coeffs_v150, kernel=self.kernel, fft=True, spin=1, compensate=True)
        ]

        # Rescale window function (by roughly a factor Ngal/Nrand in each footprint).
        w = np.zeros(3)
        w[0] = np.sum(gweights) / self.sum_rcat_gweights
        w[1] = w[2] = np.sum(vweights) / self.sum_rcat_vr_weights
        wf = self.window_function * w[:,None] * w[None,:]

        # Estimate power spectra. and normalize by dividing by window function.
        pk = core.estimate_power_spectrum(self.box, fourier_space_maps, self.kbin_edges)
        pk /= wf[:,:,None]

        # Save 'pk_data.npy' to disk. Note that the file format is specified here:
        # https://kszx.readthedocs.io/en/latest/kszpipe.html#kszpipe-details
        
        io_utils.write_npy(self.pk_data_filename, pk)
        
        return pk


    def get_pk_surrogate(self, isurr, run=False, force=False):
        r"""Returns a shape (6,6,nkbins) array, and saves it in ``pipeline_outdir/tmp/pk_surr_{isurr}.npy``.
        
        The returned array contains auto and cross power spectra of the following fields, for a single surrogate:
        
          - 0: surrogate galaxy field $S_g$ with $f_{NL}=0$.
          - 1: derivative $dS_g/df_{NL}$.
          - 2: surrogate kSZ velocity reconstruction $S_v^{90}$, with $b_v=0$ (i.e. noise only).
          - 3: derivative $dS_v^{90}/db_v$.
          - 4: surrogate kSZ velocity reconstruction $S_v^{150}$, with $b_v=0$ (i.e. noise only).
          - 5: derivative $dS_v^{150}/db_v$.
        
        Flags:

          - If ``run=False``, then this function expects the $P(k)$ file to be on disk from a previous pipeline run.
          - If ``run=True``, then the $P(k)$ file will be computed if it is not on disk.
          - If ``force=True``, then this function recomputes $P(k)$, even if it is on disk from a previous pipeline run.
        """

        fname = self.pk_single_surr_filenames[isurr]
        
        if (not force) and os.path.exists(fname):
            return io_utils.read_npy(fname)

        if not (run or force):
            raise RuntimeError(f'KszPipe.get_pk_surrogate2(): run=False was specified, and file {fname} not found')

        print(f'get_pk_surrogate({isurr}): running\n', end='')

        zobs = self.rcat.zobs
        nrand = self.rcat.size
        rweights = getattr(self.rcat, 'weight_gal', np.ones(nrand))
        vweights = getattr(self.rcat, 'weight_vr', np.ones(nrand))

        # The SurrogateFactory simulates LSS fields (delta, phi, vr) sampled at random catalog locations.
        self.surrogate_factory.simulate_surrogate()
        ngal = self.surrogate_factory.ngal

        # Noise realization for Sg (see overleaf).
        eta_rms = np.sqrt((nrand/ngal) - (self.surr_bg**2 * self.surrogate_factory.sigma2) * self.surrogate_factory.D**2)
        if np.min(eta_rms) < 0:
            raise RuntimeError('Noise RMS went negative! This is probably a symptom of not enough randoms (note {(ngal/nrand)=})')
        eta = np.random.normal(scale = eta_rms)

        # Each surrogate field is a sum (with coefficients) over the random catalog.
        # First we compute the coefficient arrays (6 fields total).
        # For more info, see the overleaf, or the sphinx docs:
        #  https://kszx.readthedocs.io/en/latest/kszpipe.html#kszpipe-details
        
        Sg = (ngal/nrand) * rweights * (self.surr_bg * self.surrogate_factory.delta + eta)
        dSg_dfnl = (ngal/nrand) * rweights * (2 * self.deltac) * (self.surr_bg-1) * self.surrogate_factory.phi
        Sv90_noise = vweights * self.surrogate_factory.M * self.rcat.tcmb_90
        Sv150_noise = vweights * self.surrogate_factory.M * self.rcat.tcmb_150
        Sv90_signal = (ngal/nrand) * vweights * self.rcat.bv_90 * self.surrogate_factory.vr
        Sv150_signal = (ngal/nrand) * vweights * self.rcat.bv_150 * self.surrogate_factory.vr

        # Mean subtraction for the surrogate field Sg.
        # This is intended to make the surrogate field more similar to the galaxy overdensity delta_g
        # which satisfies "integral constraints" since the random z-distribution is inferred from the
        # galaxies. (In practice, the effect seems to be small.)
        
        if self.nzbins_gal > 0:
            Sg = utils.subtract_binned_means(Sg, zobs, self.nzbins_gal)
            dSg_dfnl = utils.subtract_binned_means(dSg_dfnl, zobs, self.nzbins_gal)

        # Mean subtraction for the surrogate fields Sv.
        # This is intended to mitgate foregrounds.( Note that we perform the same
        # mean subtraction to the vr arrays, in get_pk_data()).
        
        if self.nzbins_vr > 0:
            Sv90_noise = utils.subtract_binned_means(Sv90_noise, zobs, self.nzbins_vr)
            Sv90_signal = utils.subtract_binned_means(Sv90_signal, zobs, self.nzbins_vr)
            Sv150_noise = utils.subtract_binned_means(Sv150_noise, zobs, self.nzbins_vr)
            Sv150_signal = utils.subtract_binned_means(Sv150_signal, zobs, self.nzbins_vr)

        # (Coefficient arrays) -> (Fourier-space fields).
        # Note spin=1 on the vr FFTs.
        
        fourier_space_maps = [
            core.grid_points(self.box, self.rcat_xyz_obs, Sg, kernel=self.kernel, fft=True, spin=0, compensate=True),
            core.grid_points(self.box, self.rcat_xyz_obs, dSg_dfnl, kernel=self.kernel, fft=True, spin=0, compensate=True),
            core.grid_points(self.box, self.rcat_xyz_obs, Sv90_noise, kernel=self.kernel, fft=True, spin=1, compensate=True),
            core.grid_points(self.box, self.rcat_xyz_obs, Sv90_signal, kernel=self.kernel, fft=True, spin=1, compensate=True),
            core.grid_points(self.box, self.rcat_xyz_obs, Sv150_noise, kernel=self.kernel, fft=True, spin=1, compensate=True),
            core.grid_points(self.box, self.rcat_xyz_obs, Sv150_signal, kernel=self.kernel, fft=True, spin=1, compensate=True)
        ]

        # Rescale window function, by a factor (ngal/nrand) in each footprint.
        wf = (ngal/nrand)**2 * self.window_function

        # Expand window function from shape (3,3) to shape (6,6).
        wf = wf[[0,0,1,1,2,2],:][:,[0,0,1,1,2,2]]

        # Estimate power spectra. and normalize by dividing by window function.
        pk = core.estimate_power_spectrum(self.box, fourier_space_maps, self.kbin_edges)
        pk /= wf[:,:,None]

        io_utils.write_npy(fname, pk)
        return pk


    def get_pk_surrogates(self):
        """Returns a shape (nsurr,6,6,nkins) array, and saves it in ``pipeline_outdir/pk_surrogates.npy``.
        
        The returned array contains auto and cross power spectra of the following fields, for all surrogates:

          - 0: surrogate galaxy field $S_g$ with $f_{NL}=0$.
          - 1: derivative $dS_g/df_{NL}$.
          - 2: surrogate kSZ velocity reconstruction $S_v^{90}$, with $b_v=0$ (i.e. noise only).
          - 3: derivative $dS_v^{90}/db_v$.
          - 4: surrogate kSZ velocity reconstruction $S_v^{150}$, with $b_v=0$ (i.e. noise only).
          - 5: derivative $dS_v^{150}/db_v$.

        This function only reads files from disk -- it does not run the pipeline.
        To run the pipeline, use :meth:`~kszx.KszPipe.run()`.
        """
        
        if os.path.exists(self.pk_surr_filename):
            return io_utils.read_npy(self.pk_surr_filename)

        if not all(os.path.exists(f) for f in self.pk_single_surr_filenames):
            raise RuntimeError(f'KszPipe.read_pk_surrogates(): necessary files do not exist; you need to call KszPipe.run()')

        pk = np.array([ io_utils.read_npy(f) for f in self.pk_single_surr_filenames ])

        # Save 'pk_surrogates.npy' to disk. Note that the file format is specified here:
        # https://kszx.readthedocs.io/en/latest/kszpipe.html#kszpipe-details
        
        io_utils.write_npy(self.pk_surr_filename, pk)
        
        return pk

        
    def run(self, processes):
        """Runs pipeline and saves results to disk, skipping results already on disk from previous runs.

        Implementation: creates a multiprocessing Pool, and calls :meth:`~kszx.KszPipe.get_pk_data()`
        and :meth:`~kszx.KszPipe.get_pk_surrogates()` in worker processes.

        Can be run from the command line with::
        
           python -m kszx kszpipe_run [-p NUM_PROCESSES] <input_dir> <output_dir>

        The ``processes`` argument is the number of worker processes. Currently I don't have a good way
        of setting this automatically -- the caller must adjust the number of processes, based on the
        size of the datasets, and amount of memory available.
        """
        
        # Copy yaml file from input to output dir.
        if not os.path.exists(f'{self.output_dir}/params.yml'):
            shutil.copyfile(f'{self.input_dir}/params.yml', f'{self.output_dir}/params.yml')
                
        have_data = os.path.exists(self.pk_data_filename)
        have_surr = os.path.exists(self.pk_surr_filename)
        missing_surrs = [ ] if have_surr else [ i for (i,f) in enumerate(self.pk_single_surr_filenames) if not os.path.exists(f) ]

        if (not have_surr) and (len(missing_surrs) == 0):
            self.get_pk_surrogates()   # creates "top-level" file
            have_surr = True
            
        if have_data and have_surr:
            print(f'KszPipe.run(): pipeline has already been run, exiting early')
            return
        
        # Initialize window function and SurrogateFactory before creating multiprocessing Pool.
        self.window_function
        self.surrogate_factory

        # FIXME currently I don't have a good way of setting the number of processes automatically --
        # caller must adjust the number of processes to the amount of memory in the node.
        
        with utils.Pool(processes) as pool:
            l = [ ]
                
            if not have_data:
                l += [ pool.apply_async(self.get_pk_data, (True,False)) ]   # (run,force)=(True,False)
            for i in missing_surrs:
                l += [ pool.apply_async(self.get_pk_surrogate, (i,True)) ]  # (run,force)=(True,False)

            for x in l:
                x.get()

        if not have_surr:
            # Consolidates all surrogates into one file
            self.get_pk_surrogates()


####################################################################################################


class KszPipeOutdir:
    def __init__(self, dirname, nsurr=None):
        r"""A helper class for loading and processing output files from ``class KszPipe``.

        Note: for MCMCs and parameter fits, there is a separate class :class:`~kszx.PgvLikelihood`.
        The KszPipeOutdir class is more minimal (the main use case is plot scripts!)

        The constructor reads the files ``{dirname}/params.yml``, ``{dirname}/pk_data.npy``,
        ``{dirname}/pk_surrogates.npy`` which are generated by :meth:`~kszx.KszPipe.run()`.
        For more info on these files, and documentation of file formats, see the sphinx docs:

           https://kszx.readthedocs.io/en/latest/kszpipe.html#kszpipe-details
          
        Constructor arguments:

          - ``dirname`` (string): name of pipeline output directory.
         
          - ``nsurr`` (integer or None): this is a hack for running on an incomplete
            pipeline. If specified, then ``{dirname}/pk_surr.npy`` is not read.
            Instead we read files of the form ``{dirname}/tmp/pk_surr_{i}.npy``.
        """

        filename = f'{dirname}/params.yml'
        print(f'Reading {filename}')
        
        with open(filename, 'r') as f:
            params = yaml.safe_load(f)

        kmax = params['kmax']
        nkbins = params['nkbins']
        surr_bg = params['surr_bg']
        kbin_edges = np.linspace(0, kmax, nkbins+1)
        kbin_centers = (kbin_edges[1:] + kbin_edges[:-1]) / 2.

        pk_data = io_utils.read_npy(f'{dirname}/pk_data.npy')
        if pk_data.shape != (3,3,nkbins):
            raise RuntimeError(f'Got {pk_data.shape=}, expected (3,3,nkbins) where {nkbins=}')

        if nsurr is None:
            pk_surr = io_utils.read_npy(f'{dirname}/pk_surrogates.npy')
            if (pk_surr.ndim != 4) or (pk_surr.shape[1:] != (6,6,nkbins)):
                raise RuntimeError(f'Got {pk_surr.shape=}, expected (nsurr,6,6,nkbins) where {nkbins=}')

        else:
            pk_surr = [ ]
            
            for i in range(nsurr):
                pk_surrogate = io_utils.read_npy(f'{dirname}/tmp/pk_surr_{i}.npy')
                if pk_surrogate.shape != (6,6,nkbins):
                    raise RuntimeError(f'Got {pk_surrogate.shape=}, expected (6,6,nkbins) where {nkbins=}')
                pk_surr.append(pk_surrogate)
                
            pk_surr = np.array(pk_surr)
            pk_surr = np.reshape(pk_surr, (nsurr,6,6,nkbins))   # needed if nsurr==0

        self.k = kbin_centers
        self.nkbins = nkbins
        self.kmax = kmax
        self.dk = kmax / nkbins
        
        self.pk_data = pk_data
        self.pk_surr = np.array(pk_surr)
        self.nsurr = self.pk_surr.shape[0]
        self.surr_bg = surr_bg


    def pgg_data(self):
        r"""Returns shape ``(nkbins,)`` array containing $P_{gg}^{data}(k)$."""
        return self.pk_data[0,0,:]
        
    def pgg_mean(self, fnl=0):
        r"""Returns shape ``(nkbins,)`` array, containing $\langle P_{gg}^{surr}(k) \rangle$."""
        assert self.nsurr >= 1
        return np.mean(self._pgg_surr(fnl), axis=0)

    def pgg_rms(self, fnl=0):
        r"""Returns shape ``(nkbins,)`` array, containing sqrt(Var($P_{gg}^{surr}(k)$))."""
        assert self.nsurr >= 2
        return np.sqrt(np.var(self._pgg_surr(fnl), axis=0))

    
    def pgv_data(self, field):
        r"""Returns shape ``(nkbins,)`` array containing $P_{gv}^{data}(k)$.

        The ``field`` argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - ``field=[1,0]`` for 90 GHz reconstruction
           - ``field=[0,1]`` for 150 GHz reconstruction
           - ``field=[1,-1]`` for null (90-150) GHz reconstruction.
        """
        field = self._check_field(field)
        return field[0]*self.pk_data[0,1] + field[1]*self.pk_data[0,2]
        
    def pgv_mean(self, field, fnl, bv):
        r"""Returns shape ``(nkbins,)`` array containing $\langle P_{gv}^{surr}(k) \rangle$.

        The ``field`` argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - ``field=[1,0]`` for 90 GHz reconstruction
           - ``field=[0,1]`` for 150 GHz reconstruction
           - ``field=[1,-1]`` for null (90-150) GHz reconstruction.
        """
        assert self.nsurr >= 1
        return np.mean(self._pgv_surr(field,fnl,bv), axis=0)

    def pgv_rms(self, field, fnl, bv):
        r"""Returns shape ``(nkbins,)`` array containing sqrt(Var($P_{gv}^{surr}(k)$)).

        The ``field`` argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - ``field=[1,0]`` for 90 GHz reconstruction
           - ``field=[0,1]`` for 150 GHz reconstruction
           - ``field=[1,-1]`` for null (90-150) GHz reconstruction.
        """
        assert self.nsurr >= 2
        return np.sqrt(np.var(self._pgv_surr(field,fnl,bv), axis=0))


    def pvv_data(self, field):
        r"""Returns shape ``(nkbins,)`` array containing $P_{vv}^{data}(k)$.

        The ``field`` argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - ``field=[1,0]`` for 90 GHz reconstruction
           - ``field=[0,1]`` for 150 GHz reconstruction
           - ``field=[1,-1]`` for null (90-150) GHz reconstruction.
        """
        field = self._check_field(field)
        t = field[0]*self.pk_data[1,:] + field[1]*self.pk_data[2,:]  # shape (3,)
        return field[0]*t[1] + field[1]*t[2]
        
    def pvv_mean(self, field, bv):
        r"""Returns shape ``(nkbins,)`` array containing $\langle P_{vv}^{data}(k) \rangle$.

        The ``field`` argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - ``field=[1,0]`` for 90 GHz reconstruction
           - ``field=[0,1]`` for 150 GHz reconstruction
           - ``field=[1,-1]`` for null (90-150) GHz reconstruction.
        """
        assert self.nsurr >= 1
        return np.mean(self._pvv_surr(field,bv), axis=0)
        
    def pvv_rms(self, field, bv):
        r"""Returns shape ``(nkbins,)`` array containing sqrt(var($P_{vv}^{data}(k)$)).

        The ``field`` argument is a length-2 vector, selecting a linear combination
        of the 90+150 GHz velocity reconstructions. For example:
        
           - ``field=[1,0]`` for 90 GHz reconstruction
           - ``field=[0,1]`` for 150 GHz reconstruction
           - ``field=[1,-1]`` for null (90-150) GHz reconstruction.
        """        
        assert self.nsurr >= 2
        return np.sqrt(np.var(self._pvv_surr(field,bv), axis=0))


    def _check_field(self, field):
        """Checks that 'field' is a 1-d array of length 2."""
        field = np.array(field, dtype=float)
        if field.shape != (2,):
            raise RuntimeError(f"Expected 'field' argument to be a 1-d array of length 2, got {field.shape=}")
        return field
    
    def _pgg_surr(self, fnl=0):
        """Returns shape (nsurr, nkbins) array, containing P_{gg} for each surrogate"""
        pgg = self.pk_surr[:,:2,:2,:]             # shape (nsurr, 2, 2, nkbins)
        pgg = pgg[:,0,:,:] + fnl * pgg[:,1,:,:]   # shape (nsurr, 2, nkbins)
        return pgg[:,0,:] + fnl * pgg[:,1,:]      # shape (nsurr, nkbins)
    
    def _pgv_surr(self, field, fnl, bv):
        """Returns shape (nsurr, nkbins) array, containing P_{gv} for each surrogate"""
        field = self._check_field(field)
        pgv = self.pk_surr[:,:2,2:,:]                                # shape (nsurr, 2, 4, nkbins)
        pgv = pgv[:,0,:,:] + fnl * pgv[:,1,:,:]                      # shape (nsurr, 4, nkbins)
        pgv = field[0] * pgv[:,0:2,:] + field[1] * pgv[:,2:4,:]      # shape (nsurr, 2, nkbins)
        return pgv[:,0,:] + bv * pgv[:,1,:]                          # shape (nsurr, nkbins)

    def _pvv_surr(self, field, bv):
        """Returns shape (nsurr, nkbins) array, containing P_{vv} for each surrogate"""
        field = self._check_field(field)
        pvv = self.pk_surr[:,2:,2:,:]                                  # shape (nsurr, 4, 4, nkbins)
        pvv = field[0] * pvv[:,0:2,:,:] + field[1] * pvv[:,2:4,:,:]    # shape (nsurr, 2, 4, nkbins)
        pvv = field[0] * pvv[:,:,0:2,:] + field[1] * pvv[:,:,2:4,:]    # shape (nsurr, 2, 2, nkbins)
        pvv = pvv[:,0,:,:] + bv * pvv[:,1,:,:]                         # shape (nsurr, 2, nkbins)
        return pvv[:,0,:] + bv * pvv[:,1,:]                            # shape (nsurr, nkbins)



from . import cpp_kernels

# Core functions.
from .core import \
    fft_r2c, \
    fft_c2r, \
    interpolate_points, \
    grid_points, \
    apply_kernel_compensation, \
    multiply_rfunc, \
    multiply_kfunc, \
    multiply_r_component, \
    apply_partial_derivative, \
    simulate_white_noise, \
    simulate_gaussian_field, \
    estimate_power_spectrum, \
    kbin_average, \
    fkp_from_ivar_2d, \
    ivar_combine, \
    estimate_cl

# "Core" classes.
from .Box import Box
from .BoundingBox import BoundingBox
from .Catalog import Catalog
from .Cosmology import Cosmology, CosmologicalParams

# "High-level" classes.
from .CmbClFitter import CmbClFitter
from .KszPipe import KszPipe, KszPipeOutdir
from .PgvLikelihood import PgvLikelihood
from .RegulatedDeconvolver import RegulatedDeconvolver
from .SurrogateFactory import SurrogateFactory

# "Utility" submodules.
from . import utils
from . import io_utils
from . import healpix_utils
from . import pixell_utils
from . import wfunc_utils
from . import plot
from . import tests

# Datasets.
from . import act
from . import desi
from . import desils_lrg
from . import desils_main
from . import planck
from . import sdss

# This submodule is a place for old code to retire peacefully.
from . import retirement_home



import io   # StringIO
import numpy as np

from . import utils

from .Box import Box


class BoundingBox(Box):
    def __init__(self, points, pixsize, rpad):
        r"""Subclass of Box whose size/position are chosen to match a galaxy survey.

        Constructor args:

           - points (numpy array with shape (ngal,3)):
               Contains galaxy locations, in coordinate system with observer at origin.
               Usually obtained by calling Catalog.get_xyz(cosmo) -- see example below.

           - pixsize (float):
               Pixel side length (caller-specified units, usually Mpc).

           - rpad (float):
               Box padding in same units as 'points' (e.g. rpad=200 for 200 Mpc).

        The BoundingBoxconstructor will automatically choose the number of pixels (npix)
        and box position (cpos), so that the box contains the galaxy survey with the
        specified padding (rpad).

        For more info on Boxes, see the :class:`~kszx.Box` base class docstring.

        Inherits the following members from :class:`~kszx.Box`:
        
           - ndim (integer): Number of dimensions (usually 3).
           - pixsize (float): Pixel side length in user-defined coordinates.
           - boxsize (length-ndim array): Box side lengths in user-defined coordinates (equal to npix * pixsize).
           - npix (length-ndim array): Real-space map shape, represented as numpy array.
           - real_space_shape (tuple): same as 'npix', but represented as tuple instead of 1-d array
           - nk (length-ndim array): Fourier-space map shape, represented as numpy array.
           - fourier_space_shape (tuple) same as 'nk', but represented as tuple instead of 1-d array
           - cpos (length-ndim array): location of box center, in observer coordinates
           - lpos (length-ndim array): location of lower left box corner, in observer coordinates
           - rpos (length-ndim array): location of upper right box corner, in observer coordinates
           - kfund (length-ndim array): Lowest frequency on each axis (equal to 2pi/boxsize)
           - knyq (float): Nyquist frequency (equal to pi/pixsize)
           - box_volume (float): Box volume (equal to prod(boxsize))
           - pixel_volume (float): Pixel volume (equal to pixsize^N)

        The BoundingBox subclass also contains the following members:

           - rpad (float): padding that was specified in BoundingBox constructor
           - rmin (float): min distance between observer and galaxies (specified in constructor)
           - rmax (float): max distance between observer and galaxies (specified in constructor)

        Example::

            # Make bounding box which contains galaxies + randoms.
            # Setup: 'gcat' and 'rcat' are objects of type Catalog.
            # 'cosmo' is an object of type Cosmology.

            gcat_xyz = gcat.get_xyz(cosmo)    # shape (ngal,3)
            rcat_xyz = rcat.get_xyz(cosmo)    # shape (rand,3)

            all xyz = np.concatenate((gcat_xyz,rcat_xyz))
            bbox = kszx.BoundingBox(all_xyz, pixsize=10, rpad=200)
        """

        points = utils.asarray(points, 'BoundingBox constructor', 'points', dtype=float)
        pixsize = float(pixsize)
        rpad = float(rpad)

        assert points.ndim ==2
        npoints, ndim = points.shape

        assert 1 <= ndim <= 3
        assert pixsize > 0
        assert rpad >= 0
        self.rpad = rpad

        r2 = np.sum(points**2, axis=1)
        self.rmin = np.min(r2)**0.5
        self.rmax = np.max(r2)**0.5
        del r2

        # These arrays are shape (3,)
        xyz_min = np.min(points, axis=0)
        xyz_max = np.max(points, axis=0)
        cpos = (xyz_min + xyz_max) / 2.
        bs = (xyz_max - xyz_min)    # boxsize, before any padding

        # These arrays are shape (3,)
        self.npix_prepad = np.array(np.ceil(bs / pixsize), dtype=int)
        self.npix_preround = np.array(np.ceil((bs+rpad) / pixsize), dtype=int)
        npix = 4 * self.npix_preround  # sentinel

        # Round up to "FFT-friendly" npix values.
        # FIXME (low-priority): move this code to its own function in kszx.utils.
        
        for p in [ 1,3,5,9,15,27 ]:
            q = (self.npix_preround - 0.5) / p
            r = np.ceil(np.log2(q))
            r = np.array(r, dtype=int)
            npix = np.minimum(npix, p * 2**r)

        Box.__init__(self, npix, pixsize, cpos=cpos)

        
    def __str__(self):
        with io.StringIO() as f:
            print(f'BoundingBox(', file=f)
            self._print_box_members(f, end='')
            print(f'    rpad = {self.rpad:.02f},', file=f)
            print(f'    rmin = {self.rmin:.02f}, rmax = {self.rmax:.02f},', file=f)
            print(f'    npix_prepad = {self.npix_prepad},', file=f)
            print(f'    npix_preround = {self.npix_preround}', file=f)
            print(f')', file=f)
            return f.getvalue()



import numpy as np

from scipy.ndimage import gaussian_filter1d
from scipy.interpolate import interp1d


class RichardsonLucyDeconvolver:
    def __init__(self, zobs_arr, zerr_arr, zmin=0.0, zmax=1.5, zerr_min=0.02, zerr_max=0.5, nzbins=100, nzerrbins=49, niter=100, sigma=2):
        r"""Photo-z model obtained by Richardson-Lucy deconvolution.

        Source: Selim's ``pipeline_getsurrogates_selim.ipynb``, Nov 2024.

        Constructor args:

          - ``zobs_arr``: 1-d array of observed redshifts.
          - ``zerr_arr``: 1-d array of observed redshift errors.
          - ``zmin, zmax, nzbins``: used internally when doing Richardson-Lucy deconvolution. 
            The defaults (0, 1.5, 100) are appropriate for DESILS-LRG.
          - ``zerr_min, zerr_max, nzerrbins``: used internally to split up the data before doing Richardson-Lucy deconvolution.
            The defaults (0.02, 0.5, 49) are approriate for DESILS-LRG.
          - ``niter``: number of Richardson-Lucy iterations.
          - ``sigma``: adjust for different smoothing levels.
        """
        
        zobs_arr = np.asarray(zobs_arr)
        zerr_arr = np.asarray(zerr_arr)
        
        # Argument checking.
        assert zobs_arr.shape == zerr_arr.shape
        assert zobs_arr.ndim == 1
        assert 0 <= zmin < zmax
        assert 0 <= zerr_min < zerr_max
        assert nzbins > 0
        assert nzerrbins > 0
        assert niter > 0
        assert sigma > 0.0

        self.nzbins = nzbins
        self.nzerrbins = nzerrbins
        self.z = np.linspace(zmin, zmax, nzbins)
        self.pzebins = np.linspace(zerr_min, zerr_max, nzerrbins+1)
        self.meanphotz = np.zeros(nzerrbins+2)
        self.umat = np.zeros((nzerrbins+2, nzbins))
        
        locatepze = np.digitize(zerr_arr, self.pzebins)

        for iz in range(nzerrbins+2):
            ix_bin = (locatepze == iz)
            zobs_bin = zobs_arr[ix_bin]
            zerr_bin = zerr_arr[ix_bin]
    
            if len(zerr_bin) < 5:
                continue
    
            self.meanphotz[iz] = np.mean(zerr_bin)
            counts_obs, bin_edges = np.histogram(zobs_bin, density=False, bins=nzbins, range=(zmin,zmax))
    
            zs_obs = (bin_edges[1:]+bin_edges[:-1])/2

            counts_obs = gaussian_filter1d(counts_obs, sigma=sigma)*1.
            counts_obs /= np.sum(counts_obs)
    
            counts_obs_int = interp1d(zs_obs, counts_obs, fill_value=0, bounds_error=False)
            observed_dist = counts_obs_int(self.z)

            Pij = np.array([[np.exp(-0.5*((zi-zj)/self.meanphotz[iz])**2) for zi in self.z] for zj in self.z])
            for j in range(len(self.z)):
                Pij[:,j] /= np.sum(Pij[:,j])

            di = observed_dist
            uguess = np.ones(nzbins)
    
            for iteration in range(niter):
                ci = np.dot(Pij, uguess)
                uguess *= np.dot(di / (ci+1.0e-10), Pij)

            self.umat[iz,:] = uguess / np.sum(uguess)

    
    def sample_z(self, zerr, zmin=0.4, zmax=1.0):
        r"""Randomly sample (ztrue, zobs) pairs, conditioned on a specified array of 'zerr' values.

        Function arguments:

          - ``zerr`` (array): 1-d array of zerr values.
          - ``(zmin, zmax)``: the returned zobs values will be constrained to satisfy 
            $z_{\rm min} < z_{\rm obs} < z_{\rm max}$. The defaults (0.4, 1.0) are appropriate
            for DESILS-LRG.

        Return values:

          - ``ztrue``: 1-d array (same length as ``zerr`` argument).
          - ``zobs``: 1-d array (same length as ``zerr`` argument).

        Each returned $(z_{\rm true}, z_{\rm obs})$ pair is consistent with the corresponding
        caller-specified $z_{\rm err}$.
        """
        
        zerr = np.asarray(zerr)
        assert zerr.ndim == 1

        locatepze = np.digitize(zerr, self.pzebins)
        ztrue = np.zeros_like(zerr)
        zobs = np.zeros_like(zerr)

        for iz in range(self.nzerrbins + 2):
            ix_bin = (locatepze == iz)
            lentruez = np.sum(ix_bin)
            
            buffer = 4
            lenpze = lentruez * buffer
    
            # Define some oversampled z-bins.
            # KMS: changed number of bins from lenpze to (10 * self.nzbins)
            # (speeds things up and gives similar results)
            
            zbin_edges = np.linspace(self.z[0], self.z[-1], 10*self.nzbins + 1)
            dz = zbin_edges[1] - zbin_edges[0]

            interp = interp1d(self.z, self.umat[iz,:])
            zvecs = (zbin_edges[1:] + zbin_edges[:-1]) / 2.  # bin centers
            zintrp = np.maximum(interp(zvecs), 0)

            # We perturb true_z within its bin [-dz/2,dz/2]
            # (Otherwise, all true_z's are at bin centers, and the histogram looks weird)
            
            true_z_sample = np.random.choice(zvecs, size=lenpze, p = (zintrp / np.sum(zintrp)))
            true_z_sample += np.random.uniform(-dz/2, dz/2, size=lenpze)
            photoz_err = np.random.normal(scale = self.meanphotz[iz], size = lenpze)
            observed_z_sample = true_z_sample + photoz_err 

            selectz_ = (observed_z_sample > zmin) * (observed_z_sample < zmax)
            ztrue[ix_bin] = (true_z_sample[selectz_])[:lentruez]
            zobs[ix_bin] = (observed_z_sample[selectz_])[:lentruez]

        return ztrue, zobs



"""
The ``kszx.act`` module contains functions for downloading/parsing ACT data products.

References: 
  - https://lambda.gsfc.nasa.gov/product/act/index.html
  - https://lambda.gsfc.nasa.gov/product/act/actadv_prod_table.html (DR5)
  - https://lambda.gsfc.nasa.gov/product/act/act_dr6.02 (DR6)
  - https://portal.nersc.gov/project/act (cluster masks from https://arxiv.org/abs/2307.01258)
"""

import os
import zipfile
import numpy as np
import pixell.enmap

from . import io_utils


def read_cmb(freq, dr, *, night=False, download=False):
    r"""Returns a pixell map. We currently only support act+planck srcfree maps.

    Function args:
    
      - ``freq`` (integer): either 90, 150, or 220.
      - ``dr`` (integer): currently, ``dr=5`` and ``dr=6`` are supported.
      - ``night`` (boolean): either True (for night) or False (for daynight).
      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.

    (Note that for ACT DR6, there are two types of ACT+Planck coadds: DR4+DR6+Planck and
    DR6+Planck. According to Mat, "I would probably go with the latter since it's more homogenous,
    and the dr4 data isn't going to make a huge difference in sensitivity", so I'm using DR6+Planck.)
    """
    
    filename = _cmb_filename(freq, dr, night=night, download=download, dlfunc='kszx.act.read_cmb')
    return _read_map(filename)


def read_ivar(freq, dr, *, night=False, download=False):
    r"""Returns a pixell map. We currently only support act+planck srcfree maps.

    Function args:
    
      - ``freq`` (integer): either 90, 150, or 220.
      - ``dr`` (integer): currently, only ``dr=5`` and ``dr=6`` are supported.
      - ``night`` (boolean): either True (for night) or False (for daynight).
      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.
    """
    
    filename = _ivar_filename(freq, dr, night=night, download=download, dlfunc='kszx.act.read_ivar')
    return _read_map(filename)


def read_beam(freq, dr, lmax=None, *, night=False, download=False):
    r"""Returns a 1-d numpy array of length (lmax+1). We currently only support act+planck srcfree maps.

    Function args:
    
      - ``freq`` (integer): either 90, 150, or 220.
      - ``dr`` (integer): currently, only ``dr=5`` and ``dr=6`` is supported.
      - ``lmax`` (integer): if None, then a large lmax will be used.
      - ``night`` (boolean): either True (for night) or False (for daynight).
      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.
    """

    filename = _beam_filename(freq, dr, night=night, download=download, dlfunc='kszx.act.read_beam')
    
    print(f'Reading {filename}\n', end='')
    a = np.loadtxt(filename)

    if dr == 5:
        assert (a.ndim == 2) and (a.shape[1] == 2)
        assert np.all(a[:, 0] == np.arange(len(a)))
        a = a[:, 1]
    elif dr == 6:
        a = a[:, 1] / a[0, 1]

    if lmax is not None:
        assert lmax < len(a)
        a = a[:(lmax+1)]
        
    return a


def read_cluster_mask(download=False):
    r"""Returns the cluster mask from https://arxiv.org/abs/2307.01258 as a pixell map."""
    
    filename = _cluster_mask_filename(download, dlfunc='kszx.act.read_cluster_mask')
    print(f'Reading {filename}\n', end='')
    return pixell.enmap.read_map(filename)


def read_nilc_wide_mask(download=False):
    r"""Returns the Galactic mask from https://arxiv.org/abs/2307.01258 as a pixell map."""
    
    filename = _nilc_wide_mask_filename(download, dlfunc='kszx.act.read_nilc_wide_mask')
    print(f'Reading {filename}\n', end='')
    return pixell.enmap.read_map(filename)


def download(dr, freq_list=None, night=False, cmb=True, ivar=True, beams=True):
    r"""Downloads ACT data products (cmb, ivar, beam) for a given survey.
        
    Can be called from command line: ``python -m kszx download_act``."""

    if freq_list is None:
        freq_list = [ 90, 150, 220 ]

    for freq in freq_list:
        _beam_filename(freq, dr, night=night, download=beams)
        _cmb_filename(freq, dr, night=night, download=cmb)
        _ivar_filename(freq, dr, night=night, download=ivar)


####################################################################################################
    
                     
def _act_path(relpath, dr, download=False, aux_relpath=None, dlfunc=None):
    """Intended to be called through wrapper such as _cmb_filename(), _ivar_filename(), etc.

    Examples (note that directory structure is a little different for DR5/DR6):

      _act_path(relpath = 'act_planck_dr5.01_s08s18_AA_f090_daynight_ivar.fits', dr=5)
      _act_path(relpath = 'maps/published/act-planck_dr6.02_coadd_AA_daynight_f090_map_srcfree.fits', dr=6)

    The 'aux_relpath' arg is used in a situation where the target file is contained in a tar.gz
    or zip file. Examples:

       aux_relpath = 'act_dr5.01_auxilliary.zip'           # DR5 example
       aux_relpath = 'beams/act_dr6.02_main_beams.tar.gz'  # DR6 example
    
    Here and in other parts of kszx, the 'dlfunc' argument gives the name of a transitive caller that
    expects the file to be present, and has a 'download=False' optional argument. This information is
    only used when generating exception-text (to tell the user how to download the file).
    """

    if dr == 5:
        dr_dir = 'dr5.01'
        url_base = 'https://lambda.gsfc.nasa.gov/data/suborbital/ACT/ACT_dr5/maps/'
    elif dr == 6:
        dr_dir = 'dr6.02'
        url_base = 'https://lambda.gsfc.nasa.gov/data/act/'
    else:
        raise RuntimeError(f'ACT {dr=} is not supported (currently support dr=5 and dr=6)')

    act_base_dir = os.path.join(io_utils.get_data_dir(), 'act', dr_dir)
    abspath = os.path.join(act_base_dir, relpath)

    if not io_utils.do_download(abspath, download, dlfunc):
        return abspath

    if aux_relpath is None:
        # Download case 1: target file is not contained in a tar.gz or zip.
        io_utils.wget(abspath, url_base + relpath)
    else:
        # Download case 2: need to download the auxfile (tar.gz or zip) first.
        # We do this by calling _act_path() recursively (with aux_relpath=None).
        aux_abspath = _act_path(aux_relpath, dr, download=download, dlfunc=dlfunc)
        io_utils.unpack(aux_abspath, expected_dstfile=abspath)  # works for .zip and .tar.
    
    return abspath


def _act_nersc_portal_path(relpath, download=False, dlfunc=None):
    """Intended to be called through wrapper such as _cluster_mask_filename().
    For files which are downloaded here: https://portal.nersc.gov/project/act
    """

    act_base_dir = os.path.join(io_utils.get_data_dir(), 'act')
    abspath = os.path.join(act_base_dir, relpath)

    if io_utils.do_download(abspath, download, dlfunc):
        url = f'https://portal.nersc.gov/project/act/{relpath}'
        io_utils.wget(abspath, url)   # calls assert os.path.exists(...) after downloading

    return abspath
    

def _cmb_filename(freq, dr, night=False, download=False, dlfunc=None):
    time = 'night' if night else 'daynight'
    
    if dr == 5:
        relpath = f'act_planck_dr5.01_s08s18_AA_f{freq:03d}_{time}_map_srcfree.fits'
    elif dr == 6:
        relpath = f'maps/published/act-planck_dr6.02_coadd_AA_{time}_f{freq:03d}_map_srcfree.fits'
    else:
        raise RuntimeError(f'ACT {dr=} is not supported (currently we support dr=5 or dr=6)')

    return _act_path(relpath, dr, download=download, dlfunc=dlfunc)


def _ivar_filename(freq, dr, night=False, download=False, dlfunc=None):
    time = 'night' if night else 'daynight'

    if dr == 5:
        relpath = f'act_planck_dr5.01_s08s18_AA_f{freq:03d}_{time}_ivar.fits'
    elif dr == 6:
        relpath = f'maps/published/act-planck_dr6.02_coadd_AA_{time}_f{freq:03d}_ivar.fits'
    else:
        raise RuntimeError(f'ACT {dr=} is not supported (currently we support dr=5 or dr=6)')
    
    return _act_path(relpath, dr, download=download, dlfunc=dlfunc)


def _beam_filename(freq, dr, night=False, download=False, dlfunc=None):
    time = 'night' if night else 'daynight'
    if dr == 5:
        relpath = f'beams/act_planck_dr5.01_s08s18_f{freq:03d}_{time}_beam.txt'
        aux_relpath = 'act_dr5.01_auxilliary.zip'
    elif dr == 6:
        # According to https://lambda.gsfc.nasa.gov/product/act/act_dr6.02/act_dr6.02_maps_info.html:
        #   Coadd maps are convolved with the coadd_{array}_{freq}_night_beam_tform_jitter_cmb.txt beams,
        #   where "array" is "pa5" for the f090 and f150 coadds and "pa4" for the f220 coadd.
        if not night: print('Note: daynight beam is not available for ACT DR6, using night beam instead.')
        adict = {90: 'pa5', 150: 'pa5', 220: 'pa4'}
        relpath = f'beams/main_beams/nominal/coadd_{adict[freq]}_f{freq:03d}_night_beam_tform_jitter_cmb.txt'
        aux_relpath = 'beams/act_dr6.02_main_beams.tar.gz'
    else:
        raise RuntimeError(f'ACT {dr=} is not supported (currently we support dr=5 or dr=6)')

    return _act_path(relpath, dr, download, aux_relpath=aux_relpath, dlfunc=dlfunc)


def _cluster_mask_filename(download=False, dlfunc=None):
    return _act_nersc_portal_path('dr6_nilc/ymaps_20230220/masks/cluster_mask.fits', download=download, dlfunc=dlfunc)


def _nilc_wide_mask_filename(download=False, dlfunc=None):
    return _act_nersc_portal_path('dr6_nilc/ymaps_20230220/masks/wide_mask_GAL070_apod_1.50_deg_wExtended.fits', download, dlfunc=dlfunc)


def _read_map(filename):
    """Called by read_cmb() and read_ivar()."""
    
    print(f'Reading {filename}\n', end='')

    # FIXME is there a way to avoid reading TQU, if we only want T?
    m = pixell.enmap.read_map(filename)
    assert m.ndim == 3
    
    # Use of .copy() saves memory, by keeping T data while dropping reference to TQU data.
    return m[0].copy()



"""
The ``kszx.wfunc_utils`` module contains functions for computing window functions.

This module is currently a placeholder -- it implements a crude approximation for the
window function, that can be used to roughly normalize $P(k)$. In the future I'll expand
it to include an accurate computation of the window function with k-bin mixing, spins, etc.
"""

import numpy as np

from . import core

from .Box import Box


def compute_wcrude(box, fourier_space_footprints, rmax=0.03):
    r"""A crude approximation to the $P(k)$ window function which neglects k-dependence and mixing.

    Given N "footprint" maps $F_{ij}(x)$, this function computes an N-by-N matrix $W_{ij}$ which
    gives the window function for a cross power spectrum on footprints i,j. (Note that since
    we're neglecting k-dependence and mixing, the window function doesn't have indices which
    correspond to k-bins.)

    For more details on what is computed, see the sphinx docs:
    
         https://kszx.readthedocs.io/en/latest/wfunc_utils.html#wcrude-details
    
    Function arguments:

       - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
         See :class:`~kszx.Box` for more info.

       - ``fourier_space_footprints`` (array or list of arrays): single or multiple Fourier-space maps.

           - Case 1: If ``fourier_space_footprints`` is an array, then it represents a single Fourier-space map.
             (The array shape should be given by ``box.fourier_space_shape`` and the dtype should be ``complex``.)
    
           - Case 2: If ``fourier_space_footprints`` is a list of arrays, then it represents multiple
             Fourier-space maps. (Each map in the list should have shape ``box.fourier_space_shape``
             and dtype ``complex``.)
       
       - ``rmax`` (float): if correlation between any two footprints is < rmax, then an exception
         will be thrown. To disable this check, set ``rmax=0``.

    The return value is either a scalar in case 1 (single Fourier-space map), or an array of shape
    ``(nmaps, nmaps)`` in case 2 (multiple maps).

    Sometimes a footprint is defined by a random catalog. Here is a reminder of how to make a
    Fourier-space map from a random catalog::

      box = ...       # instance of class kszx.Box
      cosmo = ...     # instance of class kszx.Cosmology
      randcat = ...   # instance of class kszx.Catalog
      weights = ...   # 1-d array of length randcat.size
    
      xyz = randcat.get_xyz(cosmo, zcol_name='z')
      footprint = kszx.grid_points(box, xyz, rweights, kernel='cubic', fft=True, compensate=True)
    """

    map_list, multi_map_flag = core._parse_map_or_maps(box, fourier_space_footprints, 'kszx.window.wcrude')
    nfootprints = len(map_list)

    # FIXME revisit the issue of choosing K!
    K = 0.6 * box.knyq
    edges = np.array([ 0, K / 2**(1./box.ndim), K ])

    # use_dc=True is important here!!
    pk, counts = core.estimate_power_spectrum(box, map_list, edges, use_dc=True, return_counts=True)
    assert pk.shape == (nfootprints, nfootprints, 2)
    assert counts.shape == (2,)

    # Note factor (1/V_box), which normalizes sum_k -> int d^3k/(2pi)^3
    wcrude = (pk[:,:,0] - pk[:,:,1]) * counts[0] / box.box_volume
    
    for i in range(nfootprints):
        for j in range(i+1):
            r = wcrude[i,j] / np.sqrt(wcrude[i,i] * wcrude[j,j])
            if np.abs(r) < rmax:
                raise RuntimeError(f'kszx.window.wcrude(): correlation between footprints ({j,i}) is below threshold'
                                   + f' ({r=}, {rmax=}). This probably indicates an error somewhere. This check can be'
                                   + f' disabled entirely by passing rmax=0 to wcrude().')

    return wcrude if multi_map_flag else wcrude[0,0]


def compare_pk(pk1, pk2, noisy=True):
    """A utility function I wrote for testing: compares two P(k) arrays in a normalization-independent way.

    Function arguments:

      - ``pk1``, ``pk2``: arrays of either shape (nkbins,) or (nmaps,nmaps,nkbins).

    Returns a dimensionless number which is << 1 if the P(k) arrays are nearly equal.
    """
    
    pk1 = np.asarray(pk1)
    pk2 = np.asarray(pk2)

    if pk1.shape != pk2.shape:
        raise RuntimeError(f'Got {pk1.shape=} and {pk2.shape=}, expected equal shapes')

    if pk1.ndim == 1:
        pk1 = pk1.reshape((1,1,len(pk1)))
        pk2 = pk1.reshape((1,1,len(pk2)))
    
    elif (pk1.ndim != 3) or (pk1.shape[0] != pk1.shape[1]):
        raise RuntimeError(f'Got {pk1.shape=} and {pk2.shape=}, expected (nmaps,nmaps,nkbins)')

    nmaps, nkbins = pk1.shape[0], pk1.shape[2]
    w = np.ones((nmaps, nkbins))  # weights for power spectrum comparison

    for i in range(nmaps):
        if not np.all(pk1[i,i,:] > 0):
            raise RuntimeError(f'Not all auto power spectra pk1[{i},{i},:] were > 0')
        if not np.all(pk2[i,i,:] > 0):
            raise RuntimeError(f'Not all auto power spectra pk2[{i},{i},:] were > 0')
        w[i,:] = np.sqrt(pk1[i,i,:] + pk2[i,i,:])
    
    delta = np.abs(pk1-pk2) / (w[:,None,:] * w[None,:,:])

    if noisy and (nmaps > 1):
        print('compare_pk(): pairwise epsilon_max values')
        print(np.max(delta,axis=2))

    if noisy:
        print(f'compare_pk(): global epsilon_max value: {np.max(delta)}')
    
    return np.max(delta)



from .. import Box

import numpy as np

    
def compare_arrays(arr1, arr2):
    assert arr1.shape == arr2.shape
    assert arr1.dtype == arr2.dtype

    t = arr1-arr2
    num = np.vdot(t,t)
    den = np.vdot(arr1,arr1) + np.vdot(arr2,arr2)
    
    return np.sqrt(num/den) if (den > 0.0) else 0.0


def generate_indices(shape):
    if len(shape) == 1:
        for i in range(shape[0]):
            yield (i,)
    else:
        for t in generate_indices(shape[:-1]):
            for i in range(shape[-1]):
                yield t + (i,)


def random_shape(ndim=None, nmin=1):
    if ndim is None:
        ndim = np.random.randint(1,4)
    
    ret = np.zeros(ndim, dtype=int)
    nmax = int(10000 ** (1./ndim))
    assert nmax >= (nmin+1)
    
    for d in range(ndim):
        if np.random.uniform() < 0.1:
            ret[d] = nmin
        elif np.random.uniform() < 0.1:
            ret[d] = nmin+1
        else:
            ret[d] = np.random.randint(nmin, nmax+1)

    return ret


def random_box(ndim=None, nmin=2):
    npix = random_shape(ndim, nmin)
    pixsize = np.random.uniform(1.0, 10.0)
    
    t = np.random.uniform(0, pixsize)
    cpos = np.random.uniform(-t*npix, t*npix, size=len(npix))
    return Box(npix, pixsize, cpos)


def random_kbin_edges(box, nbins=None):
    if nbins is None:
        nbins = np.random.randint(2, 11)
    
    kmax = 1.1 * np.sqrt(box.ndim) * box.knyq
    kbin_edges = np.random.uniform(0.0, kmax, nbins+1)
    kbin_edges = np.sort(kbin_edges)
    kbin_edges += np.linspace(0.0, 1.0e-10 * kmax, nbins+1)
    kbin_edges[0] = kbin_edges[0] if (np.random.uniform() < 0.5) else 0.0
    
    return kbin_edges


def map_dot_product(box, arr1, arr2, normalize=True):
    """Returns dot product of two maps (maps can be either Fourier or real space).
    
    For two real-space arrays, the dot product is defined by:
        
       f.g = (pixel volume) sum_x f(x) g(x)       [ morally int d^nx f(x) g(x) ]

   For two Fourier-space arrays, the dot product is defined by:

       f.g = (box volume)^{-1} sum_k f(k) g(k)^*  [ morally int d^nk/(2pi)^n f(k) g(k)^* ]

    If dot products are defined with these normalizations, then:

       - kszx.fft_c2r() and kszx.fft_c2r() are adjoints (transposes)
       - kszx.interpolate_points() and kszx.grid_points() are ajdoints.

    These statements are the basis for unit tests in kszx.tests.test_lss.
    """
    
    # Case 1: real-space dot product.
    if box.is_real_space_map(arr1) and box.is_real_space_map(arr2):
        return box.pixel_volume * np.vdot(arr1, arr2)

    if (not box.is_fourier_space_map(arr1)) or (not box.is_fourier_space_map(arr2)):
        raise RuntimeError('wrong shapes/dtypes in map_dot_product()')
    
    # Case 2: Fourier-space dot product, last dimension is odd.
    if box.npix[-1] % 2:
        t = np.vdot(arr1[...,0], arr2[...,0])
        t += 2 * np.vdot(arr1[...,1:], arr2[...,1:])
        
    # Case 3: Fourier-space dot product, last dimension is even and > 2.
    elif box.npix[-1] > 2:
        t = np.vdot(arr1[...,0], arr2[...,0])
        t += np.vdot(arr1[...,-1], arr2[...,-1])
        t += 2 * np.vdot(arr1[...,1:-1], arr2[...,1:-1])
        
    # Case 4: Fourier-space dot product, last dimension == 2
    else:
        t = np.vdot(arr1, arr2)
    
    return t.real / box.box_volume



"""This file is currently extremely incomplete! Some day I'll test more utility functions."""

import numpy as np

from .. import utils


def test_contract_axis():
    print('test_contract_axis(): start')
    
    for _ in range(100):
        ndim = np.random.randint(1, 5)
        shape = np.random.randint(1, 10, size=(ndim,))
        axis = np.random.randint(0, ndim)
        arr = np.random.normal(size=shape)
        weights = np.random.normal(size=shape[axis])
        
        ret = utils.contract_axis(arr, weights, axis)

        # Alternate implementation of utils.contract_axis().
        ret2 = 0
        for i,w in enumerate(weights):
            ret2 = ret2 + w * np.take(arr, i, axis=axis)

        epsilon = np.max(np.abs(ret-ret2))
        assert epsilon < 1.0e-10

    print('test_contract_axis(): pass')
    



"""Currently, the only Planck data product that I'm using is the galmask, so there's not much here."""

import os
import fitsio
import healpy
import numpy as np

from . import io_utils

    
def read_hfi_galmask(sky_percentage, apodization=0, dtype=None, download=False):
    """Returns an nside=2048 healpix map in RING ordering and Galactic coordinates.

    Default dtype is either uint8 or float32 (depending on whether apodization > 0), 
    but this can be changed with the ``dtype`` argument.

    Allowed ``sky_percentage`` values: 20, 40, 60, 70, 80, 90, 97, 99
    Allowed ``apodization`` values: 0, 2, 5 (degrees)
    If ``download`` is True, then data files will be auto-downloaded.

    Note that we use Planck release 2, since release 3 doesn't seem to have HFI
    foreground masks (it does have other masks).
    
    To apply a Planck mask to ACT data, you'll need to rotate/pixellize the mask::

       pixell_mask = pixell.reproject.healpix2map(
          healpix_mask, shape, wcs, 
          rot='gal,equ',                # NOTE coordinate rotation!!
          method='spline', order=0)     # NOTE method='spline', not method='harm'!
    """

    assert sky_percentage in [ 20, 40, 60, 70, 80, 90, 97, 99 ]
    
    abspath = _hfi_galmask_filename(apodization, download, dlfunc='kszx.planck.read_hfi_galmask')
    col_name = f'GAL0{sky_percentage}'
    print(f'Reading {abspath} ({col_name=})\n', end='')
    
    with fitsio.FITS(abspath) as f:
        if col_name not in f[1].get_colnames():
            raise RuntimeError(f'{abspath}: column name {col_name} not found')
        mask = f[1][col_name].read()

    # Convert from Healpix NESTED (in fits file) to RING (assumed by pixell)
    mask = healpy.pixelfunc.reorder(mask, n2r=True)

    if dtype is not None:
        mask = np.asarray(mask, dtype)
    elif apodization > 0:
        # Convert from big-endian float32 (numpy dtype '>f4') to native float32
        mask = np.asarray(mask, np.float32)
        
    return mask


def download():
    r"""Downloads Planck galmasks.
    
    Can be called from command line: ``python -m kszx download_planck``."""

    for apodization in [0,2,5]:
        _hfi_galmask_filename(apodization, download=True)
    

####################################################################################################


def _hfi_galmask_filename(apodization, download=False, dlfunc=None):
    """
    Allowed apodization values: 0, 2, 5 (degrees)

    Note that we use Planck release 2, since release 3 doesn't seem to have HFI
    foreground masks (it does have other masks).
    
    Here and in other parts of kszx, the 'dlfunc' argument gives the name of a transitive caller that
    expects the file to be present, and has a 'download=False' optional argument. This information is
    only used when generating exception-text (to tell the user how to download the file).
    """

    assert apodization in [0, 2, 5]    
    relpath = f'release_2/ancillary-data/masks/HFI_Mask_GalPlane-apo{apodization}_2048_R2.00.fits'
    
    planck_base_dir = os.path.join(io_utils.get_data_dir(), 'planck')
    abspath = os.path.join(planck_base_dir, relpath)

    if io_utils.do_download(abspath, download, dlfunc):
        url = f'https://irsa.ipac.caltech.edu/data/Planck/{relpath}'
        io_utils.wget(abspath, url)
    
    return abspath



"""The ``kszx.utils`` module contains miscelleanous utilities, that didn't really fit in elsewhere."""

import os
import numpy as np
import scipy.special
import scipy.integrate
import scipy.interpolate
import multiprocessing


####################################################################################################


_curr_nthreads = os.cpu_count()


def get_nthreads():
    """Returns default number of threads, for functions which are multithreaded (e.g. scipy.fft).

    Currently equal to ``os.cpu_count()``, except for worker tasks in a multiprocessing pool.
    In that case, we divide the number of threads by the number of workers in the pool.

    Note: for this to work, you should construct the pool with :func:`kszx.utils.Pool()`, 
    not ``multiprocessing.Pool()``!
    """
    
    return _curr_nthreads


def set_nthreads(nthreads, reseed_numpy_rng=False):
    """Sets the default number of threads. Not intended to be called directly -- use with caution!

    (The intended use case for this function is in :func:`kszx.utils.Pool()`, when initializing
    worker threads. In particular, the ``reseed_numpy_rng`` arg isn't really related to threads,
    but is included for convenience in kszx.utils.Pool().)
    """

    assert nthreads > 0
    assert nthreads <= os.cpu_count()
    
    global _curr_nthreads
    _curr_nthreads = nthreads

    # FIXME more code needed here, to propagate new 'nthreads' to all libraries (e.g. blas).
    # Here's a library that looks helpful: https://github.com/joblib/threadpoolctl
    
    if reseed_numpy_rng:
        # I think this is the best way to reseed numpy's global RNG.
        # Note that the global RNG is an MT19937, with 624 bytes of state.
        ss = np.random.SeedSequence(pool_size=8)
        np.random.seed(ss.generate_state(624))

    
def Pool(processes=None, reseed_numpy_rng=True):
    """Wrapper around multiprocessing.Pool(), which reseeds the numpy RNG and calls set_nthreads() in each worker.

    This function addresses some issues with multiprocessing.Pool:

       - By default, workers inherit a copy of the parent's global RNG state, so functions
         like ``np.random.normal()`` produce the same random numbers in each worker. This can
         be a huge problem -- e.g. a Pool which makes Monte Carlo sims will generate the same
         realization multiple times. We address this by re-seeding numpy's default RNG state
         in each worker.

       - Functions which use a thread pool (e.g. :func:`kszx.core.fft_r2c()`) should use fewer 
         threads when called from workers, in order to avoid oversubscription issues.

       - (Minor.) If the number of workers (``processes`` arg) is unspecified, then I'd prefer
         to use the number of CPU cores, not the value of ``os.cpu_count()`` (which is usually twice
         the number of cores, since hyperthreading is usually enabled).

    Usage is the same as ``multiprocessing.Pool()``::

      def square(i):
          return i*i

      # NOTE: I recommend calling apply_async() in a loop, not calling map_async()!
      # This is because map_async() "swallows" exceptions until the very end, whereas
      # apply_async() aborts computation as soon as an exception is encountered.
      
      with kszx.utils.Pool() as pool:
          f = [ pool.apply_async(square, (i,)) for i in range(100) ]
          f = [ x.get() for x in f ]
    
      print('Here is a list of the first 100 squares:', f)

    Function arguments:

       - ``processes`` (int): number of worker subprocesses. If None, then we use the number
         of CPU cores (which is usually half the value of os.cpu_count()).

       - ``reseed_numpy_rng`` (boolean): if True (the default), then numpy's global RNG will 
         be re-seeded in each worker thread. You should always do this, unless you're confident
         that the workers won't use the global RNG.
    """

    if processes is None:
        # FIXME assumes that hyperthreading is enabled, so that the number of threads
        # is twice the number of cores. Is there an easy way to check this? (Could do
        # it by reading and parsing /proc/cpuinfo, but that's a pain.)
        processes = max(get_nthreads() // 2, 1)
        
    assert processes > 0
    assert processes == int(processes)
    worker_nthreads = max(get_nthreads() // int(processes), 1)
    
    print(f'Creating multiprocessing pool with {processes} worker processes, and {worker_nthreads} threads/worker.')

    if reseed_numpy_rng:
        print('Global numpy RNG will be re-seeded in each worker process.')
    else:
        print('Warning: global numpy RNG is shared between worker process! (reseed_numpy_rng=False)')
    
    initargs = (worker_nthreads, reseed_numpy_rng)   # arguments to set_nthreads() called in each worker
    return multiprocessing.Pool(processes, initializer=set_nthreads, initargs=initargs)


####################################################################################################


def ra_dec_to_xyz(ra_deg, dec_deg, r=None):
    """Converts spherical polar coordinates (ra_deg, dec_deg) to cartesian coords (x,y,z).
    
    Function arguments:
    
       - ``ra_deg`` (array): RA in degrees (not radians!)
       - ``dec_deg`` (array): DEC in degrees (not radians!)
       - ``r`` (optional array): radial coordinates (units of length)

    Return value:

       - ``xyz`` (array): new array with a length-3 axis appended.
         (E.g. if ra_deg and dec_deg have shape (m,n), then xyz has shape (m,n,3)).

    The coordinate systems are related by::

        x = r * cos(dec) * cos(ra)
        y = r * cos(dec) * sin(ra)
        z = r * sin(dec)

    Note that the radial coordinate ``r`` must be specified as a comoving distance. If you
    have redshifts instead, you should call ``Cosmology.chi()``::

        cosmo = kszx.Cosmology('planck18+bao')
        xyz = kszx.utils.ra_dec_to_xyz(ra_deg, dec_deg, cosmo.chi(z=z))
    """
    
    ra_deg = np.asarray(ra_deg)
    dec_deg = np.asarray(dec_deg)
    assert ra_deg.shape == dec_deg.shape

    # FIXME memory-inefficent (uses too many temp arrays)
    
    ra = ra_deg * (np.pi/180.)
    dec = dec_deg * (np.pi/180.)
    cos_dec = np.cos(dec)
    
    ret = np.empty(ra_deg.shape + (3,), dtype=float)
    ret[...,0] = cos_dec * np.cos(ra)
    ret[...,1] = cos_dec * np.sin(ra)
    ret[...,2] = np.sin(dec)

    if r is not None:
        r = np.asarray(r)
        assert r.shape == ra_deg.shape
        ret *= np.reshape(r, r.shape + (1,))
    
    return ret


def xyz_to_ra_dec(xyz, return_r=False):
    """Converts cartesian coords (x,y,z) to spherical polar coordinates (ra_deg, dec_deg).

    Function arguments:
    
       - ``xyz`` (array): array whose last axis has length 3.
       - ``return_r`` (boolean): indicates whether radial coords are returned, see below.

    Return value:

       - ``ra_deg`` (array): RA in degrees (not radians!)
       - ``dec_deg`` (array): DEC in degrees (not radians!)
       - ``r`` (array): radial coordinates (in same units as ``xyz``)
       - If ``return_r`` is True, then the return value is a triple ``(ra_deg, dec_deg, r)``.
         If ``return_r`` is False, then the return value is a pair ``(ra_deg, dec_deg)``.

    The returned arrays have one lower dimension than ``xyz``. (E.g. if ``xyz`` has shape (m,n,3)
    then ``ra_deg``, ``dec_deg``, and ``r`` all have shape (m,n).)

    The coordinate systems are related by::

        x = r * cos(dec) * cos(ra)
        y = r * cos(dec) * sin(ra)
        z = r * sin(dec)
    """

    xyz = np.asarray(xyz)
    assert xyz.shape[-1] == 3

    x = xyz[...,0]
    y = xyz[...,1]
    z = xyz[...,2]

    ra_deg = np.arctan2(y, x)
    ra_deg *= (180./np.pi)
    
    dec_deg = np.arctan2(z, np.sqrt(x*x + y*y))
    dec_deg *= (180./np.pi)

    if return_r:
        r = np.sqrt(x*x + y*y + z*z)
        return ra_deg, dec_deg, r
    else:
        return ra_deg, dec_deg
    

####################################################################################################


def W_tophat(x):
    r"""Returns Fourier transform of a 3-d tophat W(x), where x=kR. Vectorized.
    
    W(x) is given by any of the equivalent forms:

    .. math::

       W(x) &= 3 (\sin(x) - x \cos(x)) / x^3 \\
            &= 3 j_1(x) / x \\
            &= j_0(x) + j_2(x)
    """

    # Timing showed that this implementation was fastest.
    mask = (x != 0)   # to regulate division by zero
    ret = 3 * scipy.special.spherical_jn(1,x) / np.where(mask,x,1.0)
    return np.where(mask, ret, 1.0)


####################################################################################################


def asarray(x, caller, arg, dtype=None, allow_none=False):
    if allow_none and (x is None):
        return x
    
    try:
        assert x is not None
        return np.asarray(x, dtype)
    except:
        pass

    s = f' with dtype {dtype}' if (dtype is not None) else ''
    raise RuntimeError(f"{caller}: couldn't convert {arg} to array{s} (value={x})")
    

def scattered_add(lhs, ix, rhs, normalize_sum = None):
    """Returns sum of RHS values (before applying 'normalize_sum').
    FIXME: Is there a numpy function that does this? If not, write a C++ kernel.
    """
    
    assert lhs.ndim == 1
    assert ix.ndim == 1

    n = len(ix)
    rhs = np.asarray(rhs) if (rhs is not None) else None

    # Split lhs into (scalar, vector) components (the vector can be None).
    if rhs is None:
        rs, rv = 1.0, None
    elif rhs.ndim == 0:
        rs, rv = float(rhs), None
    elif rhs.shape == (n,):
        rs, rv = 1.0, rhs
    else:
        raise RuntimeError("kszx.utils.scattered_add(): 'rhs' and 'ix' have inconsistent shapes")

    rsum = (rs*np.sum(rv)) if (rv is not None) else (rs*n)
    
    if normalize_sum:
        assert rsum > 0.0
        rs *= (normalize_sum / rsum)
        
    # FIXME slow non-vectorized loops, write C++ helper function at some point.
    # (I don't think there is a numpy function that helps.)

    if rv is None:
        for i in range(n):
            lhs[ix[i]] += rs
    elif rs == 1.0:
        for i in range(n):
            lhs[ix[i]] += rv[i]
    else:
        for i in range(n):
            lhs[ix[i]] += rs * rv[i]

    return rsum
            

####################################################################################################


def quad(f, xmin, xmax, epsabs=0.0, epsrel=1.0e-4, points=None):
    """Wrapper for scipy.integrate.quad()."""
    
    assert xmin < xmax

    if points is not None:
        # Remove points outside range (not sure if scipy does this automatically)
        points = [ x for x in points if (xmin < x < xmax) ]                                       
    
    return scipy.integrate.quad(f, xmin, xmax, epsabs=epsabs, epsrel=epsrel)[0]


def dblquad(f, xmin, xmax, ymin, ymax, *, epsabs=0.0, epsrel=1.0e-4):
    """
    Wrapper for scipy.integrate.dblquad().
      - The 'f' argument is f(x,y) [not f(y,x) as in scipy.integrate.dblquad]
      - The 'ymin' and 'ymax' arguments can either be floats or functions of x
    """

    ff = lambda x,y: f(y,x)  # swap
    return scipy.integrate.dblquad(f, xmin, xmax, ymin, ymax, epsabs=epsabs, epsrel=epsrel)[0]


def spline1d(xvec, yvec):
    """Wrapper for scipy.interpolate.InterpolatedUnivariateSpline()."""
    
    assert xvec.shape == yvec.shape
    assert xvec.ndim == yvec.ndim == 1
    return scipy.interpolate.InterpolatedUnivariateSpline(xvec, yvec)


def spline2d(xvec, yvec, zmat):
    """
    Wrapper for scipy.interpolate.RectBivariateSpline().
    
    Note that RectBivariateSpline.__call__(x,y) takes a 'grid' argument:

       grid=False  means "broadcast x,y to get a set of points"
       grid=True   means "take the Cartesian product of x,y to get a grid"

    Warning: grid=True is the default! (I usually want grid=False).
    """

    assert xvec.ndim == yvec.ndim == 1
    assert zmat.shape == (len(xvec), len(yvec))
    return scipy.interpolate.RectBivariateSpline(xvec, yvec, zmat)



def logspace(xmin, xmax, n=None, dlog=None):
    """
    Returns a 1-d array of values, uniformly log-spaced over the range (xmin, xmax).
    
    The spacing can be controlled by setting either the 'n' argument (number of points)
    or 'dlog' (largest allowed spacing in log(x)).
    """

    assert 0 < xmin < xmax
    assert (n is None) or (n >= 2)

    if (n is None) and (dlog is not None):
        n = int((np.log(xmax) - np.log(xmin)) / dlog) + 2
    elif (n is None) or (dlog is not None):
        raise RuntimeError("logspace: exactly one of 'n', 'dlog' should be None")

    ret = np.exp(np.linspace(np.log(xmin), np.log(xmax), n))
    ret[0] = xmin   # get rid of roundoff error
    ret[-1] = xmax  # get rid of roundoff error
    
    return ret


def log_uniform(xmin, xmax, size=None):
    assert 0 < xmin < xmax
    return np.exp(np.random.uniform(np.log(xmin), np.log(xmax), size=size))


def covscale(m):
    """Converts covariance matrix to correlation matrix (with 1's on the diagonal)."""
    
    m = np.array(m)
    assert m.ndim == 2
    assert m.shape[0] == m.shape[1]
    assert np.all(m.diagonal() > 0.0)

    x = 1.0 / np.sqrt(m.diagonal())
    return m * x.reshape((1,-1)) * x.reshape((-1,1))


def is_perfect_square(n):
    if n < 0:
        return False
    m = int(np.round(np.sqrt(n)))
    return n == m**2


def is_sorted(x):
    assert x.ndim == 1
    assert len(x) >= 2
    return np.all(x[:-1] < x[1:])
    

def trapezoid_weights(xmin, xmax, n):
    """Returns xvec, wvec."""

    assert xmin < xmax
    assert n >= 2
    
    xvec = np.linspace(xmin, xmax, n)
    wvec = np.full(n, (xmax-xmin)/(n-1))
    wvec[0] *= 0.5
    wvec[-1] *= 0.5

    return xvec, wvec


def array_slice(arr, axis, *args):
    assert 0 <= axis < arr.ndim
    t1 = (slice(None),) * axis
    t2 = (slice(*args),)
    return a[t1+t2]
        
        
def range_checked_index_operator(source_array, index_array, default_value=None, error_message=None):
    """Returns source_array[index_array], but substitutes 'default_value' if index is out of range."""
    
    assert source_array.ndim == 1
    assert len(source_array) > 0

    n = len(source_array)
    valid = np.logical_and(index_array >= 0, index_array < n)

    if default_value is not None:
        index_array = np.where(valid, index_array, 0)
        return np.where(valid, source_array[index_array], default_value)
    
    if np.all(valid):
        return source_array[index_array]

    if error_message is None:
        error_message = 'index out of range, and no default_value specfied'

    raise RuntimeError(error_message)


def contains_duplicates(l):
    return len(l) != len(set(l))


def boxcar_sum(a, m, normalize=False):
    """
    Given a 1-d array 'a' of length n, return a length-n array containing boxcar-summed values of a.
    Each boxcar consists of (2*m+1) array elements, or fewer near the edges of the array.

    E.g. if m=2, then

      ret[0] = a[0] + a[1] + a[2]
      ret[1] = a[0] + a[1] + a[2] + a[3]
      ret[2] = a[0] + a[1] + a[2] + a[3] + a[4]
      ret[3] = a[1] + a[2] + a[3] + a[4] + a[5]
        ...
      ret[n-4] = a[n-6] + a[n-5] + a[n-4] + a[n-3] + a[n-2]
      ret[n-3] = a[n-5] + a[n-4] + a[n-3] + a[n-2] + a[n-1]
      ret[n-2] = a[n-4] + a[n-3] + a[n-2] + a[n-1]
      ret[n-1] = a[n-3] + a[n-2] + a[n-1]
    
    If 'normalize' is True, then "boxcar means" are computed instead of "boxcar sums".
    """

    # conversion to float helps with float32 dynamic-range issues
    a = np.asarray(a, dtype=float)
    assert a.ndim == 1
    assert m >= 0

    if m == 0:
        return np.copy(a)
    
    n = len(a)
    assert n > 0
        
    if n <= m+1:
        t = np.mean(a) if normalize else np.sum(a)
        return np.full((n,), t, dtype=a.dtype)

    c = np.cumsum(a)
    ret = np.zeros(n, dtype=a.dtype)
    ret[:-m] = c[m:]
    ret[-m:] = c[-1]
    ret[(m+1):] -= c[:(-m-1)]
    
    if normalize:
        den = np.full((n,), 2*m+1)
        den[:m] -= np.arange(m,0,-1)
        den[-m:] -= np.arange(1,m+1)
        ret = ret / den

    return ret


def subtract_binned_means(data, x, nbins):
    """Averages a 1-d 'data' array in x-bins, and returns a copy of 'data' with binned means subtracted."""

    assert nbins >= 1
    assert data.shape == x.shape

    # Compute bin edges
    bins = np.linspace(np.min(x), np.max(x), nbins + 1)
    
    # Digitize x values into bins
    bin_indices = np.digitize(x, bins) - 1  # subtract 1 to make bin_indices 0-based

    # Clip bin indices to ensure they're within range
    bin_indices = np.clip(bin_indices, 0, nbins - 1)

    # Compute sum and count per bin
    sum_per_bin = np.bincount(bin_indices, weights=data, minlength=nbins)
    count_per_bin = np.bincount(bin_indices, minlength=nbins)
    
    # Avoid division by zero
    count_per_bin = np.maximum(count_per_bin, 1)
    mean_per_bin = sum_per_bin / count_per_bin

    # Subtract binned means
    return data - mean_per_bin[bin_indices]


def random_rotation_matrix(N):
    """Returns a random N-by-N rotation matrix."""

    # Precise meaning of "random": drawn from Haar measure on O(N).
    # Algorithm: apply Gram-Schmidt process to a Gaussian random matrix.

    ret = np.random.normal(size=(N,N))
    
    for i in range(N):
        for j in range(i):
            c = np.dot(ret[i,:], ret[j,:])
            ret[i,:] -= c * ret[j,:]
        ret[i,:] /= np.sqrt(np.dot(ret[i,:],ret[i,:]))

    return ret


def contract_axis(arr, weights, axis):
    """'Contract a D-dimensional array with a 1-d weights array, returning a (D-1) dimensional array.

    Function arguments:

      - ``arr``: D-dimensional array of shape ``(n[0], n[1], ..., n[D-1])``.

      - ``weights`` 1-d array of length ``n[axis]``.

      - ``axis``: integer 0 <= axis < D.

    Returns a (D-1) dimensional array of shape ``(n[0], ..., n[axis-1], n[axis+1], ..., n[D-1])``,
    defined by "contracting" the weights array along the specified axis of 'arr'.
    """
    
    arr = np.asarray(arr)
    weights = np.asarray(weights)
    
    assert 0 <= axis < arr.ndim
    assert weights.shape == (arr.shape[axis],)

    # Transpose 'arr' so that 'axis' is last.
    perm = np.zeros(arr.ndim, dtype=int)
    perm[:axis] = np.arange(axis)
    perm[axis:-1] = np.arange(axis+1, arr.ndim)
    perm[-1] = axis
    arr = np.transpose(arr, perm)

    # Now we can just call np.dot().
    return np.dot(arr, weights)





from .. import Box
from .. import core
from .. import utils
from . import helpers

import itertools
import numpy as np


####################################################################################################


def test_fft_inverses():
    """Tests that fft_r2c() and fft_c2r() are inverses (spin-0 only)."""

    print('test_fft_inverses(): start')
    
    for iouter in range(100):
        box = helpers.random_box()
        x = core.simulate_white_noise(box, fourier=False)
        Fx = core.fft_r2c(box, x)
        FFx = core.fft_c2r(box, Fx)
        eps = np.max(np.abs(x-FFx)) * np.sqrt(box.pixel_volume)
        assert eps < 1.0e-13
        
    print('test_fft_inverses(): pass')


def test_fft_transposes():
    """Tests that fft_r2c() and fft_c2r() are transposes (for arbitrary spin)."""
    
    print('test_fft_transposes(): start')

    for iouter in range(100):
        box = helpers.random_box()
        spin = 0 if np.all(box.npix <= 2) else np.random.randint(0,2)

        x = core.simulate_white_noise(box, fourier=False)
        y = core.simulate_white_noise(box, fourier=True)
        Fx = core.fft_r2c(box, x, spin=spin)
        Fy = core.fft_c2r(box, y, spin=spin)

        mdot = lambda v,w: helpers.map_dot_product(box,v,w)
        dot1 = mdot(y,Fx)
        dot2 = mdot(Fy,x)
        den = mdot(y,y)*mdot(Fx,Fx) + mdot(Fx,Fx)*mdot(y,y)
        epsilon = np.abs(dot1-dot2) / den**0.5
        assert epsilon < 1.0e-12
        
    print('test_fft_transposes(): pass')
    

####################################################################################################


class RandomPoly:
    def __init__(self, degree, lpos, rpos):
        assert lpos.shape == rpos.shape
        assert lpos.ndim == 1
        assert degree >= 1

        self.ndim = len(lpos)
        self.degree = degree
        self.lpos = lpos
        self.rpos = rpos
        self.cpos = (rpos + lpos) / 2.
        self.coeffs = np.zeros((self.ndim, degree+1))

        for axis in range(self.ndim):
            dx = (rpos[axis] - lpos[axis]) / 2.
            for i in range(degree+1):
                self.coeffs[axis,i] = np.random.uniform(-1.0, 1.0) / dx**i


    def _eval_axis(self, x, axis):
        t = x - self.cpos[axis]
        tpow = np.ones_like(t)
        ret = np.zeros_like(t)

        for i in range(self.degree+1):
            ret += self.coeffs[axis,i] * tpow
            tpow *= t

        return ret
    
            
    def eval_points(self, points):
        assert points.ndim == 2
        assert points.shape[-1] == self.ndim
        
        npoints = points.shape[0]
        ret = np.ones(npoints)

        for axis in range(self.ndim):
            ret *= self._eval_axis(points[:,axis], axis)

        return ret
        
        
    def eval_grid(self, shape, ishift):
        assert len(shape) == len(ishift) == self.ndim
        ret = np.ones(shape)

        for axis in range(self.ndim):
            x = np.linspace(self.lpos[axis], self.rpos[axis], shape[axis])
            x = np.roll(x, ishift[axis])
            y = self._eval_axis(x, axis)
            s = np.concatenate((np.ones(axis,dtype=int), (-1,), np.ones(self.ndim-axis-1,dtype=int)))
            ret *= np.reshape(y, s)

        return ret


####################################################################################################


def test_interpolation():
    print('test_interpolation(): start')

    for _ in range(100):
        # Currently, interpolate_points() supports CIC and cubic.
        kernel, degree = ('cic',1) if (np.random.uniform() < 0.5) else ('cubic',3)
        periodic = (np.random.uniform() < 0.5)
        
        # Currently, interpolate_points() only supports ndim=3.
        box = helpers.random_box(ndim=3, nmin=degree+1)
        ndim = box.ndim  # placeholder for future expansion
        
        ishift = np.random.randint(-1000,1000,size=ndim) if periodic else np.zeros(ndim,dtype=int)
        poly_lpos = box.lpos + ishift * box.pixsize
        poly_rpos = box.rpos + ishift * box.pixsize
        
        poly = RandomPoly(degree=degree, lpos=poly_lpos, rpos=poly_rpos)
        
        npoints = np.random.randint(100, 200)
        pad = (degree - 1 + 1.0e-7) * (box.pixsize/2.)
        points = np.random.uniform(poly_lpos + pad, poly_rpos - pad, size=(npoints,ndim))
        exact_vals = poly.eval_points(points)

        grid = poly.eval_grid(box.npix, ishift)
        interpolated_vals = core.interpolate_points(box, grid, points, kernel=kernel, periodic=periodic)

        epsilon = helpers.compare_arrays(exact_vals, interpolated_vals)
        # print(f'{epsilon=}')
        assert epsilon < 1.0e-12
        
    print('test_interpolation(): pass')


def test_interpolation_gridding_consistency():
    print('test_interpolation_gridding_consistency(): start')
    
    for _ in range(100):
        # Currently, interpolate_points() supports CIC and cubic.
        kernel, degree = ('cic',1) if (np.random.uniform() < 0.5) else ('cubic',3)
        periodic = (np.random.uniform() < 0.5)
        
        # Currently, interpolate_points() only supports ndim=3.
        box = helpers.random_box(ndim=3, nmin=degree+1)
        ndim = box.ndim  # placeholder for future expansion

        npoints = np.random.randint(100, 200)
        pad = (-1000 * box.pixsize) if periodic else ((degree - 1 + 1.0e-7) * (box.pixsize/2.))
        points = np.random.uniform(box.lpos+pad, box.rpos-pad, size=(npoints,ndim))

        g = np.random.normal(size=box.npix)  # random grid
        wscal = np.random.uniform(1.0, 2.0)  # random wscal

        # w = random weights, either 0-d or 1-d (passed to core.grid_points() below)
        # w1 = 1-d representation of 'w' (passed to np.dot() below)
        if np.random.uniform() < 0.5:
            w = w1 = np.random.normal(size=npoints)
        else:
            w = np.random.uniform(1.0, 2.0)
            w1 = np.full(npoints, w)
        
        Ag = wscal * core.interpolate_points(box, g, points, kernel=kernel, periodic=periodic)
        Aw = core.grid_points(box, points, weights=w, kernel=kernel, periodic=periodic, wscal=wscal)

        dot1 = np.dot(w1,Ag)
        dot2 = helpers.map_dot_product(box,Aw,g)
        
        den = np.dot(w1,w1) * np.dot(Ag,Ag)
        den += helpers.map_dot_product(box,g,g) * helpers.map_dot_product(box,Aw,Aw)

        epsilon = np.abs(dot1-dot2) / den**(0.5)
        assert epsilon < 1.0e-12
        
    print('test_interpolation_gridding_consistency(): pass')


####################################################################################################


# FIXME rename this
def test_simulate_gaussian():
    print('test_simulate_gaussian(): start')
    
    for _ in range(100):
        box = helpers.random_box()
        arr = core.simulate_white_noise(box, fourier=True)
        arr2 = core.fft_c2r(box, arr)
        arr2 = core.fft_r2c(box, arr2)
        eps = np.max(np.abs(arr-arr2)) *  np.sqrt(box.pixel_volume)
        assert eps < 1.0e-10
        
    print('test_simulate_gaussian(): pass')


# FIXME turn into proper unit test
def monte_carlo_simulate_gaussian(npix, pixsize):
    box = Box(npix, pixsize)
    p_real = np.zeros(box.fourier_space_shape)
    p_fourier = np.zeros(box.fourier_space_shape)
        
    for nmc in itertools.count():
        x = core.simulate_white_noise(box, fourier=False)
        x = core.fft_r2c(box, x)
        p_real += np.abs(x)**2

        y = core.simulate_white_noise(box, fourier=True)
        p_fourier += np.abs(y)**2

        if utils.is_perfect_square(nmc):
            r = p_fourier/p_real
            print(f'{nmc=} min_ratio={np.min(r)} max_ratio={np.max(r)}')
            
        
####################################################################################################


def _test_estimate_power_spectrum(box, kbin_edges, nmaps=None, use_dc=False):
    """Compares estimate_power_spectrum() to a reference implementation.
    
    Note that nmaps=1 is slightly different from nmaps=None. 
    (The pk arrays have shape (1,1,nkbins) versus (nkbins,) in the two cases.)

    In principle, there is a small chance that this test can fail, if estimate_power_spectrum() 
    and the reference implementation put a k-mode in different bins, due to roundoff error. 
    """

    M = nmaps if (nmaps is not None) else 1
    maps = np.array([ core.simulate_white_noise(box, fourier=True) for _ in range(M) ])
    
    map_arg = maps if (nmaps is not None) else maps[0]
    pk, bc = core.estimate_power_spectrum(box, map_arg, kbin_edges, use_dc=use_dc, allow_empty_bins=True, return_counts=True)
    
    nbins = len(kbin_edges) - 1
    ref_pk = np.zeros((M,M,nbins))
    ref_bc = np.zeros(nbins, dtype=int)

    for ix in helpers.generate_indices(box.fourier_space_shape):
        i = np.array(ix, dtype=int)
            
        if (not use_dc) and np.all(i==0):
            continue

        k = np.array(i, dtype=float)
        k = np.minimum(k, box.npix - k)
        k *= (2*np.pi) / box.boxsize
        k = np.dot(k,k)**0.5
        b = np.searchsorted(kbin_edges, k, side='right')-1
        
        if (b < 0) or (b >= nbins):
            continue

        t = 1 if ((ix[-1]==0) or (2*ix[-1] == box.npix[-1])) else 2
        z = maps[(slice(None),) + ix]
        assert z.shape == (M,)
        
        ref_pk[:,:,b] += t * np.outer(z,z.conj()).real
        ref_bc[b] += t

    for b in range(nbins):
        if ref_bc[b] > 0:
            ref_pk[:,:,b] /= (ref_bc[b] * box.box_volume)

    if nmaps is None:
        ref_pk = ref_pk[0,0,:]
    
    tol = 1.0e-10 * (bc + ref_bc)**0.5
    
    assert np.all(bc == ref_bc)
    assert np.all(np.abs(pk - ref_pk) <= tol)


def test_estimate_power_spectrum():
    """Compares estimate_power_spectrum() to a reference implementation.
    
    In principle, there is a small chance that this test can fail, if estimate_power_spectrum() 
    and the reference implementation put a k-mode in different bins, due to roundoff error. 
    """

    print('test_estimate_power_spectrum(): start')
    
    for iouter in range(100):
        box = helpers.random_box()
        kbin_edges = helpers.random_kbin_edges(box)
        use_dc = (np.random.uniform() < 0.5)

        # Note: C++ kernel calls different functions for M=1, M=2, M=3, M=4, and M>4.
        nmaps = np.random.randint(1,8) if (np.random.uniform() < 0.95) else None
        # print(f'{nmaps=}')

        _test_estimate_power_spectrum(box, kbin_edges, nmaps=nmaps, use_dc=use_dc)
        
    print('test_estimate_power_spectrum(): pass')


####################################################################################################


def _test_kbin_average(box, kbin_edges, use_dc=False):
    """Compares kbin_average() to a reference implementation.
    
    In principle, there is a small chance that this test can fail, if kbin_average() 
    and the reference implementation put a k-mode in different bins, due to roundoff error. 
    """

    # Random function f(k) = sum_i coeffs[i] * cos(rvals[i]*k)
    ncoeffs = 5
    rmax = 2.0 / np.min(box.kfund)
    rvals = np.random.uniform(low=0, high=rmax, size=ncoeffs)
    coeffs = np.random.normal(size=ncoeffs)

    def f(k):
        ret = np.zeros_like(k)
        for i in range(ncoeffs):
            ret += coeffs[i] * np.cos(rvals[i]*k)
        return ret

    fmean, bc = core.kbin_average(box, f, kbin_edges, use_dc=use_dc, allow_empty_bins=True, return_counts=True)

    nbins = len(kbin_edges) - 1
    ref_fk = f(box.get_k())
    ref_fmean = np.zeros(nbins)
    ref_bc = np.zeros(nbins, dtype=int)

    for ix in helpers.generate_indices(box.fourier_space_shape):
        i = np.array(ix, dtype=int)
            
        if (not use_dc) and np.all(i==0):
            continue

        k = np.array(i, dtype=float)
        k = np.minimum(k, box.npix - k)
        k *= (2*np.pi) / box.boxsize
        k = np.dot(k,k)**0.5
        b = np.searchsorted(kbin_edges, k, side='right')-1
        
        if (b < 0) or (b >= nbins):
            continue

        t = 1 if ((ix[-1]==0) or (2*ix[-1] == box.npix[-1])) else 2
        ref_fmean[b] += t * ref_fk[ix]
        ref_bc[b] += t

    for b in range(nbins):
        if ref_bc[b] > 0:
            ref_fmean[b] /= ref_bc[b]

    eps = np.max(np.abs(fmean - ref_fmean))
    # print(f'{eps=}')
    assert np.all(bc == ref_bc)
    assert eps < 1.0e-12
    

def test_kbin_average():
    """Compares kbin_average() to a reference implementation.
    
    In principle, there is a small chance that this test can fail, if kbin_average() 
    and the reference implementation put a k-mode in different bins, due to roundoff error. 
    """

    print('test_kbin_average(): start')
    
    for iouter in range(100):
        box = helpers.random_box()
        kbin_edges = helpers.random_kbin_edges(box)
        use_dc = (np.random.uniform() < 0.5)
        _test_kbin_average(box, kbin_edges, use_dc=use_dc)
        
    print('test_kbin_average(): pass')
